/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2010 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "pim/AppointmentBuilderV85.h"
#include "base/Log.h"
#include "base/stringUtils.h"
#include "pim/difftime.h"
#include "pim/UtilsRecurrence.h"
#include "pim/MAPIUtils.h"
#include "base/util/utils.h"
#include "syncml/core/PropParam.h"

/**
 * Array of appointment fields names for SIF format (for compatibility, version 8.5)
 * Recurring fields are listed in recurrenceFields[] array.
 * @note Don't change the order of these properties!
 *       This list MUST be synchronized with PIMPR_props_app array of CEPROPID fields.
 */
const wchar_t* appointmentFieldsV85[] = {
    {L"Subject"                            },                // Returns or sets the subject for the Outlook item. This property corresponds to the MAPI property PR_SUBJECT. The Subject property is the default property for Outlook items.
    {L"AllDayEvent"                        },                // True if the appointment is an all-day event (as opposed to a specified time). Corresponds to the All day event check box on the Appointment page of an AppointmentItem - MUST be set AFTER "Start" and "End" fields.
    {L"Sensitivity"                        },                // Returns or sets the sensitivity for the Outlook item. Can be one of the following OlSensitivity constants: olConfidential(3), olNormal(0), olPersonal(1), or olPrivate(2). This property corresponds to the MAPI property PR_SENSITIVITY
    {L"IsRecurring"                        },                // True if the appointment or task is a recurring appointment or task. When the GetRecurrencePattern method is used with an AppointmentItem or TaskItem object, this property is set to True
    {L"ReminderSet"                        },                // True if a reminder has been set for this appointment, mail item or task. 
//  {L"ReminderSoundFile"                  },                // Returns or sets the path and filename of the sound file to play when the reminder occurs for the appointment or task. This property is only valid if the ReminderOverrideDefault and ReminderPlaySound properties are set to True
    {L"ReminderMinutesBeforeStart"         },                // Returns or sets the number of minutes the reminder should occur prior to the start of the appointment
//  {L"Duration"                           },                // The length of an appointment in minutes.
    {L"Location"                           },                // Returns or sets the specific office location (for example, Building 1 Room 1 or Suite 123) for the appointment. This property corresponds to the MAPI property PR_OFFICE_LOCATION
    {L"Start"                              },                // Returns or sets the starting date and time for the appointment or journal entry. Use only in calendar
    {L"End"                                },                // Returns or sets the end date and time of an appointment or journal entry. Use only on calendar
    {L"BusyStatus"                         },                // Returns or sets the busy status of the user for the appointment. Can be one of the following OlBusyStatus constants: olBusy(2), olFree(0), olOutOfOffice(3), or olTentative(1). 
    {L"ReminderOptions"                    },                // What actions to take when the reminder occurs. (LED=1, Vibrate=2, Dialog=4, Sound=8, Repeat=16)
    {L"Timezone"                           },                // The timezone of the appointment as a TIME_ZONE_INFORMATION structure
    {L"MeetingStatus"                      },                // OlRemoteStatus can be one of these OlRemoteStatus constants.
//  {L"MeetingOrganizerName"               },                // The name of the person who organized the meeting.
//  {L"MeetingOwnerCriticalChange"         },                // A timestamp indicating that the meeting owner wants to update the meeting request. Required when sending an updated meeting request to attendees.
//  {L"MeetingAttendeesCriticalChange"     },                // A timestamp indicating that a meeting attendee wants to update the meeting request. Required when sending an updated meeting request back to the meeting owner.
    {L"Categories"                         },                // Returns or sets the categories assigned to the Outlook item. 
    {L"Body"                               },                // Returns or sets the clear-text body of the Outlook item.
    {NULL}
};



/**
 * Enumeration of appointment fields used by WM plugin.
 * @note Don't change the order of these properties!
 *       This list MUST be synchronized with PIMPR_props_app array of CEPROPID fields.
 */
enum {
    ePIMPR_SUBJECT                           ,
    ePIMPR_ALL_DAY_EVENT                     ,
    ePIMPR_SENSITIVITY                       ,
    ePIMPR_IS_RECURRING                      ,
    ePIMPR_REMINDER_SET                      ,
//  ePIMPR_REMINDER_SOUND_FILE               ,
    ePIMPR_REMINDER_MINUTES_BEFORE_START     ,
//  ePIMPR_DURATION                          ,
    ePIMPR_LOCATION                          ,
    ePIMPR_START                             ,
    ePIMPR_END                               ,
    ePIMPR_BUSY_STATUS                       ,    
    ePIMPR_REMINDER_OPTIONS                  ,
//  ePIMPR_GLOBAL_OBJECT_ID                  ,
    ePIMPR_TIMEZONE                          ,
    ePIMPR_MEETING_STATUS                    ,      // <- this is removed after v8.5
//  ePIMPR_MEETING_ORGANIZER_NAME            ,
//  ePIMPR_MEETING_OWNER_CRITICAL_CHANGE     ,
//  ePIMPR_ATTENDEES_CRITICAL_CHANGE         ,
    ePIMPR_CATEGORIES                        ,
    ePIMPR_BODY_TEXT                         ,
    
    // --- Recurring fields from here ---
    ePIMPR_RECURRING_TYPE                    ,
    ePIMPR_RECURRING_PATTERNSTARTDATE        ,
    ePIMPR_RECURRING_PATTERNENDDATE          ,
    ePIMPR_RECURRING_STARTTIME               ,
    ePIMPR_RECURRING_ENDTIME                 ,
    ePIMPR_RECURRING_NOEND                   ,
    ePIMPR_RECURRING_OCCURRENCES             ,
    ePIMPR_RECURRING_INTERVAL                ,
    ePIMPR_RECURRING_DAYOFWEEKMASK           ,
    ePIMPR_RECURRING_DAYOFMONTH              ,
    ePIMPR_RECURRING_INSTANCE                ,
//  ePIMPR_RECURRING_DURATION                ,
    ePIMPR_RECURRING_MONTHOFYEAR             ,

    // This is the total number of appointment properties supported (including recurring ones)
    NUM_COLS_APP
};


/**
 * Array of appointment fields used by WM plugin.
 * **********************************************
 *
 * Recurring properties are appended at the end.
 * @note Don't change the order of these properties!
 */
static CEPROPID PIMPR_props_app[] = {    
    PIMPR_SUBJECT                           ,
    PIMPR_ALL_DAY_EVENT                     ,
    PIMPR_SENSITIVITY                       ,
    PIMPR_IS_RECURRING                      ,
    PIMPR_REMINDER_SET                      ,
//  PIMPR_REMINDER_SOUND_FILE               ,
    PIMPR_REMINDER_MINUTES_BEFORE_START     ,    
//  PIMPR_DURATION                          ,
    PIMPR_LOCATION                          ,
    PIMPR_START                             ,
    PIMPR_END                               ,
    PIMPR_BUSY_STATUS                       ,
    PIMPR_REMINDER_OPTIONS                  ,
//  PIMPR_GLOBAL_OBJECT_ID                  ,
    PIMPR_TIMEZONE                          ,
    PIMPR_MEETING_STATUS                    ,           // <- this is removed after v8.5
//  PIMPR_MEETING_ORGANIZER_NAME            ,
//  PIMPR_MEETING_OWNER_CRITICAL_CHANGE     ,
//  PIMPR_ATTENDEES_CRITICAL_CHANGE         ,
    PIMPR_CATEGORIES                        ,
    PIMPR_BODY_TEXT                         ,
    
    // --- Recurring fields from here ---
    PIMPR_RECURRING_TYPE                    ,
    PIMPR_RECURRING_PATTERNSTARTDATE        ,
    PIMPR_RECURRING_PATTERNENDDATE          ,
    PIMPR_RECURRING_STARTTIME               ,
    PIMPR_RECURRING_ENDTIME                 ,
    PIMPR_RECURRING_NOEND                   ,
    PIMPR_RECURRING_OCCURRENCES             ,
    PIMPR_RECURRING_INTERVAL                ,
    PIMPR_RECURRING_DAYOFWEEKMASK           ,
    PIMPR_RECURRING_DAYOFMONTH              ,
    PIMPR_RECURRING_INSTANCE                ,
//  PIMPR_RECURRING_DURATION                ,
    PIMPR_RECURRING_MONTHOFYEAR
};




WinItem* appointmentToWinEventV85(IItem *pAppointment, bool isSif) {

    //MessageBox (NULL, TEXT("appbuilder"), TEXT("title"), MB_SETFOREGROUND |MB_OK);

    HRESULT hr                      = E_FAIL;
    CEPROPVAL *prgPropvalPoom       = NULL;
    ULONG cbBuffer                  = 0;
    HANDLE hHeap                    = GetProcessHeap();
    WinEvent* winE                  = NULL;
    WinRecurrence* winRec           = NULL;
    IAppointment* pAppointmentOld   = NULL;
    IRecurrencePattern* pRecurrence = NULL;
    IExceptions* pExceptions        = NULL;
    int numExceptions               = 0;
    bool isAllDay                   = false;
    bool isRecurring                = false;
    wchar_t tempValue[10];
    wstring value, start;

    if (isSif) { winE = new WinEventSIF(); }
    else       { winE = new WinEvent()   ; }

    cbBuffer = 0;
    hr = pAppointment->GetProps(PIMPR_props_app, CEDB_ALLOWREALLOC, NUM_COLS_APP, &prgPropvalPoom, &cbBuffer, hHeap);
    if(FAILED(hr) || NULL == prgPropvalPoom || 0 == cbBuffer) {
        goto Exit;
    }


    //
    // Fill all appointment properties
    //
    for (int i = 0; i < ePIMPR_RECURRING_TYPE; i++) {

        switch (i)
        {
        case ePIMPR_ALL_DAY_EVENT:
            {
                int tmp = getIntPropertyFromMapi(&prgPropvalPoom[i]);
                wsprintf(tempValue, TEXT("%d"), tmp);
                value = tempValue;
                if (value == TEXT("1")) {
                    isAllDay = true;
                }
                break;
            }
        case ePIMPR_IS_RECURRING:
            {
                int tmp = getIntPropertyFromMapi(&prgPropvalPoom[i]);
                wsprintf(tempValue, TEXT("%d"), tmp);
                value = tempValue;
                if (value == TEXT("1")) {
                    isRecurring = true;
                }
                break;
            }
        case ePIMPR_START:
        case ePIMPR_END:
            {
                // Date values: "yyyyMMdd" or "YYYYMMDDThhmmssZ"
                value = getDatePropertyFromMapi(&prgPropvalPoom[i], !isAllDay, isAllDay);
                break;
            }
        case ePIMPR_SENSITIVITY:
        case ePIMPR_REMINDER_SET:        
      //case ePIMPR_DURATION:
        case ePIMPR_BUSY_STATUS:
        case ePIMPR_REMINDER_OPTIONS:
        case ePIMPR_REMINDER_MINUTES_BEFORE_START:
        case ePIMPR_MEETING_STATUS:                     // <- this is removed after v8.5
            {
                // Int values.
                int tmp = getIntPropertyFromMapi(&prgPropvalPoom[i]);
                wsprintf(tempValue, TEXT("%d"), tmp);
                value = tempValue;
                break;
            }                    
        case ePIMPR_TIMEZONE:
            {
                // Timezone: if set, recurring props are sent in local time.
                // We pass the tz structure to API.
                TIME_ZONE_INFORMATION* tz = readTimezoneFromMapi(&prgPropvalPoom[i]);
                if (tz) {
                    winE->setTimezone(tz);
                    delete [] tz;
                }
                continue;
            }
        default:
            {
                // String values.
                value = getStringPropertyFromMapi(&prgPropvalPoom[i]);
                break;
            }
        }

        winE->setProperty(appointmentFieldsV85[i], value);
        //LOG.info("%2d) %ls = \"%ls\"", i+1, appointmentFields[i], value.c_str());
    }


    if (!isRecurring) {
        goto Exit;
    }


    //
    // Manage recurring properties: need to retrieve all values together
    // ---------------------------- to apply safe checks and conversions.
    //
    winRec = winE->getRecPattern();
    if (!winRec) {
        goto Exit;
    }
    winE->getProperty(TEXT("Start"), start);
    setWinRecurrenceProps(*winRec, prgPropvalPoom, ePIMPR_RECURRING_TYPE, start, isAllDay, true);


    //
    // Exceptions. Use old style IAppointment interface.
    // -----------
    long oid = 0;
    hr = pAppointment->get_Oid(&oid);
    if (!oid || FAILED(hr))             { goto Exit; }

    IP_OUTLOOK_APP* pOutlookApp = getOutlookApp();
    if (!pOutlookApp)                   { goto Exit; }

    hr = pOutlookApp->GetItemFromOid(oid, (IDispatch**)&pAppointmentOld);
    if (!pAppointmentOld || FAILED(hr)) { goto Exit; }

    hr = pAppointmentOld->GetRecurrencePattern(&pRecurrence);
    if (!pRecurrence || FAILED(hr))    { goto Exit; }

    hr = pRecurrence->get_Exceptions(&pExceptions);
    if (!pExceptions || FAILED(hr))    { goto Exit; }

    pExceptions->get_Count(&numExceptions);
    if (numExceptions == 0) {
        goto Exit;
    }

    DATE startDate = NULL;
    systemTimeToDouble(start.c_str(), &startDate, NULL);

    // Push all EXDATEs into WinEvent list (we have only EXDATEs)
    for (int i=0; i < numExceptions; i++) {
        IException* pException;
        hr = pExceptions->Item(i + 1, &pException);
        if (!pException || FAILED(hr)) { goto Exit; }

        wstring exception = createException(pException, startDate, isAllDay);
        replaceAll(TEXT("-"), TEXT(""), exception);     // vCal: "YYYYMMDD"
        winE->getExcludeDates()->push_back(exception);
        pException->Release();
    }


Exit:
    // Release objects.
    if (pExceptions)     { pExceptions->Release();    }
    if (pRecurrence)     { pRecurrence->Release();    }
    if (pAppointmentOld) { pAppointmentOld->Release();}

    // Free memory.
    HeapFree(hHeap, 0, prgPropvalPoom);
    return winE;
}

