/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include "pim/mail/MessagingManager.h"
#include "mail/MailAccount.h"
#include "pim/mail/MailUtils.h"
#include "pim/mail/MailClient2Server.h"
#include "pim/mail/MailServer2Client.h"
#include "OutOfMemoryException.h"
#include "pim/mail/MessagingStatus.h"
#include "customization.h"

USE_NAMESPACE

MessagingManager* MessagingManager::pinstance = NULL;

/// Used for general access to mapi store/folders/messages
enum {
    eePR_ENTRYID,
    eePR_DISPLAY_NAME,
    eePR_CE_UNIQUE_STORE_ID,
    eePR_CE_TRANSPORT_NAME,
    eeNUM_COLS
};

// These tags represent the message information we would like to pick up
static SizedSPropTagArray(eeNUM_COLS, ParamsForMsgStore) =
{
    eeNUM_COLS,
    PR_ENTRYID,
    PR_DISPLAY_NAME,
    PR_CE_UNIQUE_STORE_ID,
    PR_CE_TRANSPORT_NAME,
};


// These tags represent the message information we would like to pick up
static SizedSPropTagArray(eeNUM_COLS, Columns) =
{
    eeNUM_COLS,
    PR_ENTRYID
};


/**
 * Utility, to release an unknown object that can be an account, folder or email message.
 */
static void releaseObject(LPUNKNOWN* object, unsigned long objectType) {

    if (objectType == MAPI_STORE) {
        IMsgStore* pAccount = (IMsgStore*)object;
        RELEASE_OBJ(pAccount);
    }
    else if (objectType == MAPI_FOLDER) {
        IMAPIFolder* pFolder = (IMAPIFolder*)object;
        RELEASE_OBJ(pFolder);
    }
    else if (objectType == MAPI_MESSAGE) {
        IMessage* pMessage = (IMessage*)object;
        RELEASE_OBJ(pMessage);
    }
    return;
}


/**
 * Utility, returns true if the folder's role is one of the defaults.
 */
static bool isDefaultFolderRole(const StringBuffer& role) {

    if ( (role.icmp(ROLE_INBOX)  ) ||
         (role.icmp(ROLE_OUTBOX) ) ||
         (role.icmp(ROLE_DRAFT)  ) ||
         (role.icmp(ROLE_SENT)   ) ||
         (role.icmp(ROLE_TRASH)  ) ||
         (role.icmp(ROLE_DELETED)) ) {
        return true;
    }
    return false;
}



MessagingManager::MessagingManager(MailSyncSourceConfig& ssc) : MailAccountManager(ssc), 
                                                                pSession(NULL)
{
}

MessagingManager* MessagingManager::getInstance(MailSyncSourceConfig& ssc)
{
	if (pinstance == NULL) {
		pinstance = new MessagingManager(ssc); 
	}

	return pinstance;
}

MessagingManager* getMessagingManager(/*MailSyncSourceConfig& ssc*/)
{
    MailSyncSourceConfig* ssc = getRegConfig()->getMailSyncSourceConfig(true);
	return MessagingManager::getInstance(*ssc);
}

MessagingManager::~MessagingManager() {
 
    if (pSession) {
        logoff();
	}
}


int MessagingManager::logon() {

    if (pSession != NULL) {
        LOG.debug("Already logged to IMAPISession");
        return 0;
    }

    // First log on to the store.
    HRESULT hr = MAPILogonEx (NULL, NULL, NULL, NULL, &pSession);   // returns S_OK if no error
    return hr;
}

void MessagingManager::logoff() {

    if (pSession != NULL) {
        pSession->Release();
        pSession = NULL;
    }
}

int MessagingManager::createClientAccount(MailAccount& account)
{
    int ret = NOERROR;
    wstring guid;
    wstring accountId;
    const char* accountName = account.getName();
    wchar_t* wAccountName = NULL;

    if ((accountName == NULL) || (strlen(accountName) == 0)) {
        LOG.error("%s: account name is not specified", __FUNCTION__);
        return E_FAIL;
    }

    LOG.debug("checking for account \"%s\"", accountName);
    wAccountName = toWideChar(accountName);
    
    // check if an account of that name is already on MAPI store
    if (getAccountGuidByName(wAccountName, &guid) == false) {
        return E_FAIL;    
    }

    if (guid.empty()) { // no account found: add it
        LOG.debug("account \"%s\" not found: creating it...", accountName);
 
        ret = accountProvisioner.addAccount(account, guid);
    
        if (ret) {
            LOG.error("Error in AccountProvisioner::createAccount: %i", ret);
            return ret;
        } else {
            MailSourceManagementNode m = MailSourceManagementNode(SOURCES_CONTEXT, "mails");
            MailSyncSourceConfig sc = m.getMailSourceConfig(true);

            // Retrieve the new ID from the GUID, and set it into the account.
            accountId = getAccountID(guid.c_str());
            account.setID(accountId.c_str());

            // save account in config as soon as it's created in MAPI store            
            sc.addMailAccount(account);
            m.setMailSourceConfig(sc);
        }
    } else {        
        // Retrieve the new ID from the GUID, and set it into the account.
        accountId = getAccountID(guid.c_str());
        account.setID(accountId.c_str());

        LOG.debug("account \"%s\" already created: reusing its GUID", accountName);
    }

    return ret;
}


int MessagingManager::updateClientAccount(const MailAccount& account) {

    wstring guid = getAccountGUID(account.getID());

    int ret = accountProvisioner.modAccount(account, guid.c_str());
    if (ret) {
        LOG.error("Error in AccountProvisioner::updateAccount: %i", ret);
    }
    return ret;
}


int MessagingManager::deleteClientAccount(const WCHAR* accountID){
   
    int ret = 1;

    if (accountID) {
        wstring guid = getAccountGUID(accountID);

        ret = accountProvisioner.delAccount(guid.c_str());
        if (ret) {
            LOG.error("Error in AccountProvisioner::deleteAccount: %i", ret);
        }
    }

    return ret;
}

int MessagingManager::deleteAllAccount(){
    int ret = 1;
    
    ArrayList mailAccounts = this->config.getMailAccounts();
    if (mailAccounts.size() == 0){
        ret = 0;
    }
    for(int i = 0; i < mailAccounts.size(); i++){ 
        StringBuffer idofaccount = getIdOfAccount(((MailAccount*)mailAccounts[i])->getName());
        WCHAR* wid = toWideChar(idofaccount);
        ret = deleteClientAccount(wid);
        delete [] wid; wid = NULL;
        if (ret){
            LOG.error("Error in deleteAllAccount: %i", ret);
            return ret;
        }
    }
    return ret;
}

int MessagingManager::createClientFolder(FolderData& folder) {

    // If here, target parent has already been checked.
    // Just check the role if it's ok.

    int ret = 0;
    const StringBuffer& role = folder.getRole();

    if (isDefaultFolderRole(role)) {
        // Set the folder's key, to be returned to the Server!
        if (folder.getParent().empty() == false) {
            wstring key = getIdOfFolder(folder.getParent(), role);
            folder.setID(key.c_str());
            ret = 0;
        }
        else {
            LOG.error("Cannot create folder: parent ID missing");
            ret = 2;
        }
    }
    else {
        LOG.error("Undefined folder's role: %s", role.c_str());
        ret = 1;
    }

    return ret;
}

int MessagingManager::updateClientFolder(const FolderData& folder) {
    
    // If here, target parent has already been checked.
    // Just check the role if it's ok.

    const StringBuffer& role = folder.getRole();
    if (isDefaultFolderRole(role)) {
        return 0;
    }
    else {
        LOG.error("Undefined folder's role: %s", role.c_str());
        return 1;
    }
}

int MessagingManager::deleteClientFolder(const WCHAR* folderID) {

    int ret = 0;
    unsigned long objectType = 0;
    LPUNKNOWN* object = openObject(folderID, &objectType);

    if (object) {
        if (objectType == MAPI_FOLDER) {
            // We return OK but we didn't delete the folder
            ret = 0;
        }
        else {
            LOG.error("MessagingManager::deleteClientFolder: key '%S' is not a folder item", folderID);
            ret = -1;
        }
        releaseObject(object, objectType);
    }
    else {
        LOG.error("MessagingManager::deleteClientFolder: folder '%S' not found", folderID);
        ret = -1;
    }

    return ret;
}



int MessagingManager::deleteObject(const WCHAR* key) {

    if (key == NULL) {
        return S_FALSE;
    }

    // Backward compatibility: old email IDs are "I/<key in b64>"
    WCHAR* itemKey = (WCHAR*)key;
    if ((wcslen(key) > 2) && (key[0] == 'I') && (key[1] == '/')) {
        itemKey = (WCHAR*)key+2;
    }
    
    HRESULT hr = S_OK;
    unsigned long objectType = 0;
    SBinary sb = createIdFromKey(itemKey);
    LPUNKNOWN* object = openObject(sb, &objectType);
    

    // EMAIL ACCOUNT
    if (objectType == MAPI_STORE) {
        hr = deleteClientAccount(itemKey);
        goto FuncExit;
    } 
    // EMAIL FOLDER
    else if (objectType == MAPI_FOLDER) {
        // Not necessary: we already checked the object is a folder.
        //hr = deleteClientFolder(itemKey);
        hr = S_OK;
        goto FuncExit;
    } 
    // EMAIL MESSAGE
    else if (objectType == MAPI_MESSAGE) {
        hr = deleteMessage((IMessage*)object, sb);
        goto FuncExit;
    }
    else {
        hr = S_FALSE;
        goto FuncExit;
    }

FuncExit: 
    releaseObject(object, objectType);
    return hr;
}



// Method to delete an email given its key from the server.
/*bool MessagingManager::deleteEmail(const WCHAR* key) {
    
    bool ret = true;
    SBinary mailId = createIdFromKey(key);
    IMessage* pMessage = getIMessage(key);
    
    if (deleteEmail(pMessage, mailId)) {
        ret = false;
    }

    RELEASE_OBJ(pMessage);
    return ret;
}*/

int MessagingManager::addEmail(MailClientData* m) {
    return addEmailOnMapiStore(m, false);
}

int MessagingManager::addEmailWithAttachment(MailClientData* m) {
    return addEmailOnMapiStore(m, true);
}

int MessagingManager::addEmailOnMapiStore(MailClientData* m, bool addAttachment) {

	HRESULT hr               = S_OK;
    LPUNKNOWN*   object      = NULL;
	IMAPIFolder* folder      = NULL;
	IMessage*    pmsg		 = NULL;

    ULONG        rgMsgTags[] = { 1, PR_ENTRYID };
    ULONG        cmsgValue   = 0;
    SPropValue*  msgprops    = NULL;

	unsigned long objectType = 0;
	const wchar_t* parentId  = NULL;

	if (!m) {
        LOG.error("Update message: mail client data is NULL.");
        return S_FALSE;
    }

    // some sanity check on parent ID
	if (((parentId = m->getParent()) == NULL) ||
            (wcslen(parentId) == 0)) {
		LOG.error("unable to get folder id");
		return S_FALSE;		
	}

	SBinary sb = createIdFromKey(parentId); 

    if (sb.cb == 0) {
	    LOG.error("invalid parent ID");
		return S_FALSE;
    }

	// openObject will create the MAPI session if necessary
    object = openObject(sb, &objectType);
	if (object == NULL) {
        return S_FALSE;
    }
    if (objectType != MAPI_FOLDER) {
        releaseObject(object, objectType);
		return S_FALSE;
	} 
    folder = (IMAPIFolder*)object;
                
	hr = folder->CreateMessage(NULL, 0, &pmsg);
	EXIT_ON_FAILED(hr);

	if (!pmsg){
		LOG.error("addMessage: pmsg NULL!");
        goto FuncExit;
	}

	// 
    // Get the ID of the new message
    //
	hr = pmsg->GetProps ((SPropTagArray *) rgMsgTags, MAPI_UNICODE, &cmsgValue, &msgprops);	

	try {                       
		// Now set the props on the message....
        hr = mailServer2Client.SetMessageProps (pSession, pmsg, m, addAttachment);
	} 
	
	catch (OutOfMemoryException&) {
        deleteStoredMessage(folder, msgprops);
        throw OutOfMemoryException(0, -50); 
    }

    if (hr != S_OK) {
        deleteStoredMessage(folder, msgprops);
        LOG.error("Error in insert new mail message.");
        goto FuncExit;
    }

    LOG.debug("Message props set.");
	
	if (msgprops[eePR_ENTRYID].ulPropTag == PR_ENTRYID) {
        wstring msgId = createKeyFromId(msgprops[eePR_ENTRYID].Value.bin);
		m->setEntryId(msgId.c_str());
    } else {
        LOG.error("Error getting msg entry id.");
    }

FuncExit:
	MAPIFreeBuffer (msgprops);
    RELEASE_OBJ (pmsg);
    RELEASE_OBJ (folder);

    return hr;
}

int MessagingManager::updateEmail(const MailClientData* m) {

	HRESULT hr					= S_OK;
	IMessage* msg				= NULL;
	unsigned long objectType	= 0;
	const wchar_t* msgId		= NULL;

	if (!m) {
        LOG.error("updateEmail: mail client data is NULL.");
        return S_FALSE;
    }

	if ((msgId = m->getEntryId()) == NULL) {
		LOG.error("updateEmail: unable to get entry id");
		return S_FALSE;		
	}

	SBinary sb = createIdFromKey(msgId); 
    // openObject will create the MAPI session if necessary
	if ((msg = (IMessage *)openObject(sb, &objectType)) == NULL) {  
        return S_FALSE;
    }

    if (objectType != MAPI_MESSAGE) {
		LOG.error("updateEmail: object from store is not a mail message");
        releaseObject((LPUNKNOWN*)msg, objectType);
		hr = S_FALSE;
		goto FuncExit;
	}

	hr = mailServer2Client.SetMessageProps (pSession, msg, const_cast<MailClientData *>(m)); // the decision in done by the MailMessage

FuncExit:
    RELEASE_OBJ (msg);
    
	return hr;
}

HRESULT MessagingManager::deleteStoredMessage(LPMAPIFOLDER folder, SPropValue* rgprops) {
    
    HRESULT hr = S_OK;
    if (folder == NULL || rgprops == NULL) {
        return S_FALSE;
    }

    ENTRYLIST entryList;
    entryList.cValues = 1;
    entryList.lpbin[0] = rgprops[eePR_ENTRYID].Value.bin;

    hr = folder->DeleteMessages (&entryList, 0, NULL, 0);

    return hr;
}



int MessagingManager::deleteMessage(IMessage* pMessage, SBinary mailId) {

    IMAPIFolder* folder = getParentFolder(pMessage);
    if (!folder) {
        return S_FALSE;
    }
                
    ENTRYLIST entryList;                
    entryList.cValues = 1;
    entryList.lpbin = &mailId;    
    HRESULT hr = folder->DeleteMessages(&entryList, 0, NULL, 0);

    RELEASE_OBJ(folder);
    return hr;
}


IMAPIFolder* MessagingManager::getParentFolder(IMessage* pMessage) {

    IMAPIFolder* folder = NULL;
    if (!pMessage) {
        return NULL;
    }

    ULONG mailProps[2];
    mailProps[0] = 1;
    mailProps[1] = PR_PARENT_ENTRYID;
    ULONG        cValues = 0;
    SPropValue*  rgprops = NULL;
    pMessage->GetProps((LPSPropTagArray) &mailProps, MAPI_UNICODE, &cValues, &rgprops);

    unsigned long objectType = 0;
    LPUNKNOWN* object = openObject(rgprops[0].Value.bin, &objectType);
    if (object && objectType == MAPI_FOLDER) {
        folder = (IMAPIFolder*)object;
    }
    else {
        releaseObject(object, objectType);
        return NULL;
    }

    return folder;
}


wstring MessagingManager::getEntryIdOfFolder(IMAPIFolder* folder) {

    wstring folderID;
    if (!folder) {
        return folderID;
    }

    ULONG* rgTags = new ULONG[2];
    rgTags[0] = 1;
    rgTags[1] = PR_ENTRYID;
    ULONG        cValues = 0;
    SPropValue*  rgprops = NULL;

     // Get the folder ID.
    HRESULT hr = folder->GetProps((SPropTagArray *) rgTags, MAPI_UNICODE, &cValues, &rgprops);
    EXIT_ON_FAILED (hr);
    folderID = createKeyFromId(rgprops[0].Value.bin); 

// needed by EXIT_ON_FAILED macro
FuncExit:
    return folderID;
}


wstring MessagingManager::getParentFolderID(const WCHAR* messageID) {

    IMessage*    pMsg   = NULL;
    IMAPIFolder* folder = NULL;
    wstring folderID;

    if (!messageID) {
        goto FuncExit;
    }

    pMsg = getIMessage(messageID);
    if (!pMsg) {
        LOG.error("getParentFolderID: error opening msg with ID = %S", messageID);
        goto FuncExit;
    }

    folder = getParentFolder(pMsg);
    if (!folder) {
        LOG.error("getParentFolderID: error getting parent folder for msg with ID = %S", messageID);
        goto FuncExit;
    }

    folderID = getEntryIdOfFolder(folder);
    
FuncExit:
    RELEASE_OBJ(pMsg);
    RELEASE_OBJ(folder);
    
    return folderID;
}


int MessagingManager::getAllMessages(ArrayList& mailItems, const wstring& folderId, const StringBuffer& folderRole)
{
    HRESULT         hr;
    UINT            nEntries = 0;
    LPMAPITABLE     pContentsTable = NULL;
    LPSRowSet       pRows = NULL;
    
    LPUNKNOWN*      object = NULL;
    unsigned long   objectType = 0;
    IMAPIFolder*    folder = NULL;

    SBinary sb = createIdFromKey(folderId.c_str()); 
	
    // openObject will create the MAPI session if necessary
    object = openObject(sb, &objectType);
	if (object == NULL) {
        return S_FALSE;
    }
    if (objectType != MAPI_FOLDER) {
        releaseObject(object, objectType);
		return S_FALSE;
	} 
    folder = (IMAPIFolder*)object;

    // We define a SPropTagArray array here using the SizedSPropTagArray Macro
    // This enum will allows us to access portions of the array by a name instead of a number.
    // If more tags are added to the array, appropriate constants need to be added to the enum.
    hr = folder->GetContentsTable(0, &pContentsTable);
    EXIT_ON_FAILED(hr);

    hr = pContentsTable->SetColumns((LPSPropTagArray)&Columns, 0);
    EXIT_ON_FAILED(hr)     
       
    while (!FAILED(hr = pContentsTable->QueryRows(1, 0, &pRows))) {
        if (pRows->cRows == 0)
            break;

        SBinary sbMsgEntry = pRows->aRow[0].lpProps[eePR_ENTRYID].Value.bin;
        wstring msgId = createKeyFromId(sbMsgEntry);
        
        ItemData msgItem(msgId, MAPI_MESSAGE);
        msgItem.setParentId(folderId);
        msgItem.setParentRole(folderRole);

        mailItems.add(static_cast<ItemData>(msgItem));
    }

    hr = S_OK;

FuncExit:
    if (pRows) {
        // Free each element in the pRows buffer
        FreeProws(pRows);
    }

    RELEASE_OBJ(pContentsTable);
    RELEASE_OBJ(folder);

    return hr;
}

/*unsigned long MessagingManager::getObjectType(const wchar_t* key) {
    
    // Check if it's a mail message/folder/account
    LPUNKNOWN* object = getIMessage(key);

    if (object) {
        RELEASE_OBJ(object);
        return MAPI_MESSAGE;
    }
    
    object = getIMsgStore(key);
    if (object) {
        RELEASE_OBJ(object);
        return MAPI_STORE;
    }

    object = getIMapiFolder(key);
    if (object) {
        RELEASE_OBJ(object);
        return MAPI_FOLDER;
    }
    return 0;
}*/



LPUNKNOWN* MessagingManager::openObject(const wchar_t* key, unsigned long* requestedType) {    
    // get the SBinary structure with the binary id
    SBinary sb = createIdFromKey(key);
    return openObject(sb, requestedType);
}


LPUNKNOWN* MessagingManager::openObject(SBinary sb, unsigned long* requestedType) {
    
    HRESULT hr = S_OK;
    ULONG FAR       lpulObjType;
    LPUNKNOWN FAR * lppUnk;
    
    // open the session if necessary
    if (!pSession) {
        hr = logon();
        EXIT_ON_FAILED (hr);
    }

    hr = pSession->OpenEntry (sb.cb,
                              (LPENTRYID)sb.lpb,
                              NULL,
                              NULL,
                              &lpulObjType,
                             (IUnknown **) &lppUnk);
    
    *requestedType = lpulObjType;
    return lppUnk;
   
FuncExit: 
    return NULL;
}



// return IMsgStore from the b64 key
IMsgStore* MessagingManager::getIMsgStore(const WCHAR* key) {
    
    LPUNKNOWN FAR * lppUnk  = NULL;
    IMsgStore* pStore       = NULL;
    unsigned long objectType = 0;

    lppUnk = openObject(key, &objectType);
    if (lppUnk && objectType == MAPI_STORE) {
        pStore = (IMsgStore*)lppUnk;
    }
    return pStore;
}

// return IMAPIFolder from the b64 key
IMAPIFolder* MessagingManager::getIMapiFolder(const WCHAR* key) {
    
    LPUNKNOWN FAR * lppUnk  = NULL;
    IMAPIFolder* pFolder    = NULL;
    unsigned long objectType = 0;

    lppUnk = openObject(key, &objectType);
    if (lppUnk && objectType == MAPI_FOLDER) {
        pFolder = (IMAPIFolder*)lppUnk;
    }
    return pFolder;
}

// return IMessage from the b64 key
IMessage* MessagingManager::getIMessage(const wchar_t* key) {
    
    LPUNKNOWN FAR * lppUnk  = NULL;
    IMessage* pMessage      = NULL;
    unsigned long objectType = 0;

    lppUnk = openObject(key, &objectType);
    if (lppUnk && objectType == MAPI_MESSAGE) {
        pMessage = (IMessage*)lppUnk;
    }
    return pMessage;
}



StringBuffer MessagingManager::createKeyFromIdA(const SBinary sbEntry) {
    
    StringBuffer ret;
    if (sbEntry.cb == 0 || sbEntry.lpb == NULL) {
        return ret;
    }

    int len = sbEntry.cb;
    char* b64 = new char[((len/3+1)<<2) + 1];
    len = b64_encode(b64, sbEntry.lpb, len);
    b64[len] = 0;
    ret = b64;
    delete [] b64;
    return ret;
}

wstring MessagingManager::createKeyFromId(const SBinary sbEntry) {
    
    wstring ret;    
    StringBuffer s = createKeyFromIdA(sbEntry);
    if (!s.empty()) {
        wchar_t* t = toWideChar(s.c_str());
        ret.append(t);
        delete [] t;
    }
    return ret;
}

SBinary MessagingManager::createIdFromKey(const wchar_t* entryId) {

    SBinary sbEntry;
    sbEntry.cb = 0;
    sbEntry.lpb = NULL;

    char* eid = toMultibyte(entryId);
    if (!eid) {
        LOG.error("decodeEntryId: error coverting entryId");
        return sbEntry;
    }    
    size_t len = b64_decode(eid, eid);
    sbEntry.cb = len;
    sbEntry.lpb = (BYTE*)eid;
        
    return sbEntry;
}



// Returns the account ID (key) from its GUID.
wstring MessagingManager::getAccountID(const wchar_t* guid) {

    HRESULT           hr;
    IMAPITable*       pTable      =  NULL;
    SRowSet*          psrs        =  NULL;
    IMsgStore*        pStore      =  NULL;
    wstring           ret;
    GUID*             pGuid       =  NULL;
    wstring           paramGuid(guid);

    // open the session if necessary
    if (!pSession) {
        hr = logon();
        EXIT_ON_FAILED (hr);
    }
    
    // Get the message stores table
    hr = pSession->GetMsgStoresTable (MAPI_UNICODE, &pTable);
    EXIT_ON_FAILED (hr);

    hr = pTable->SetColumns((LPSPropTagArray)&ParamsForMsgStore, 0);
    EXIT_ON_FAILED(hr);

    while (1) {
        hr = pTable->QueryRows (1, 0, &psrs);
        EXIT_ON_FAILED (hr);

        if (psrs->cRows != 1) {
          break;
        }

        // check if the property is right
        if (psrs->aRow[0].lpProps[eePR_CE_UNIQUE_STORE_ID].ulPropTag == PR_CE_UNIQUE_STORE_ID) {
            pGuid = psrs->aRow[0].lpProps[eePR_CE_UNIQUE_STORE_ID].Value.lpguid;
            wstring foundGuid = accountProvisioner.formatGUID(pGuid);
            
            if (foundGuid == paramGuid) {
                if (psrs->aRow[0].lpProps[eePR_ENTRYID].ulPropTag == PR_ENTRYID) {
                    ret = createKeyFromId(psrs->aRow[0].lpProps[eePR_ENTRYID].Value.bin);   
                    break;
                }
            }
        }
    }

FuncExit:
    return ret;
}


// Returns the account GUID from its key (ID).
wstring MessagingManager::getAccountGUID(const wchar_t* key) 
{
    wstring ret;
    HRESULT hr         = 0;
    IMAPITable* pTable =  NULL;
    SRowSet*    psrs   =  NULL;
    IMsgStore*  pStore =  NULL;

    // open the session if necessary
    if (!pSession) {
        hr = logon();
        EXIT_ON_FAILED (hr);
    }

    // Get the message stores table
    hr = pSession->GetMsgStoresTable (MAPI_UNICODE, &pTable);
    EXIT_ON_FAILED(hr);

    hr = pTable->SetColumns((LPSPropTagArray)&ParamsForMsgStore, 0);
    EXIT_ON_FAILED(hr);

    while (1) {
        // Get a row
        hr = pTable->QueryRows(1, 0, &psrs);
        EXIT_ON_FAILED(hr);

        // Did we hit the end of the table?
        if (psrs->cRows != 1) {
          break;
        }

        if (psrs->aRow[0].lpProps[eePR_ENTRYID].ulPropTag == PR_ENTRYID) {

            //wstring currentName = psrs->aRow[0].lpProps[eePR_DISPLAY_NAME].Value.lpszW;
            //if (currentName == TEXT("Funambol")) {

            wstring currentKey = createKeyFromId(psrs->aRow[0].lpProps[eePR_ENTRYID].Value.bin);
            if (currentKey == key) {

                // Account found
                LPGUID pGuid = NULL;
                if (psrs->aRow[0].lpProps[eePR_CE_UNIQUE_STORE_ID].ulPropTag == PR_CE_UNIQUE_STORE_ID) {
                    pGuid = psrs->aRow[0].lpProps[eePR_CE_UNIQUE_STORE_ID].Value.lpguid;
                }

                if (pGuid) {
                   ret = accountProvisioner.formatGUID(pGuid);
                   goto FuncExit;
                } 
            }
        }
    }
      

FuncExit:
    return ret;
}

// Returns the account GUID from it display name.
bool MessagingManager::getAccountGuidByName(const wchar_t* accountName, wstring* guid) 
{
    //wstring     guid;
    bool ret           = true;
    HRESULT     hr     = 0;
    IMAPITable* pTable = NULL;
    SRowSet*    psrs   = NULL;
    IMsgStore*  pStore = NULL;
 
    if (accountName == NULL) {
        LOG.error("%s: account name is null", __FUNCTION__);
     
        return false;
    }

    // open the session if necessary
    if (!pSession) {
        hr = logon();
        
        if (FAILED(hr)) {
            LOG.error("error in mapi log on");
            return false;
        }
    }

    // Get the message stores table
    hr = pSession->GetMsgStoresTable (MAPI_UNICODE, &pTable);
    
    if (FAILED(hr)) {
        LOG.error("error getting MAPI message store tables");
        return false;
    }

    hr = pTable->SetColumns((LPSPropTagArray)&ParamsForMsgStore, 0);
    if (FAILED(hr)) {
        LOG.error("error setting MAPI message store columns");
        return false;
    }
   
    while (1) {
        // Get a row
        hr = pTable->QueryRows (1, 0, &psrs);
        if (FAILED(hr)) {
            LOG.error("error quering MAPI message store");
            return false;
        }

        // Did we hit the end of the table?
        if (psrs->cRows != 1) {
            break;
        }

        if (psrs->aRow[0].lpProps[eePR_DISPLAY_NAME].ulPropTag == PR_DISPLAY_NAME) { 
            wchar_t* displayName = psrs->aRow[0].lpProps[eePR_DISPLAY_NAME].Value.lpszW;
            wchar_t* transportName = NULL;

            if (psrs->aRow[0].lpProps[eePR_CE_TRANSPORT_NAME].ulPropTag == PR_CE_TRANSPORT_NAME) {
                transportName = psrs->aRow[0].lpProps[eePR_CE_TRANSPORT_NAME].Value.lpszW;
            }

            if ((displayName != NULL) && (wcscmp(displayName, accountName) == 0)) {
                LPGUID pGuid = NULL;

                if (transportName == NULL) {
                    LOG.error("can't find transport name for account \"%S\"", displayName);
                  
                    return false;
                }

                // check if the account found is of type SyncML
                if (wcscmp(transportName, TEXT(SERVICE_TYPE))) {
                    LOG.error("account \"%S\" is already defined but isn't of type SyncML", accountName);
                   
                    return false;
                }

                if (psrs->aRow[0].lpProps[eePR_CE_UNIQUE_STORE_ID].ulPropTag == PR_CE_UNIQUE_STORE_ID) {
                    pGuid = psrs->aRow[0].lpProps[eePR_CE_UNIQUE_STORE_ID].Value.lpguid;
                }

                if (pGuid) {
                    *guid = accountProvisioner.formatGUID(pGuid);
                } else {
                    LOG.error("error getting GUID account from mapi query");

                    return false;
                }
               
                break;
            }
        }
    }

    return true;
}


ULONG MessagingManager::getFolderEntryId(const StringBuffer& t) {

    ULONG ret = -1;

    if (t == ROLE_INBOX)
        ret = PR_CE_IPM_INBOX_ENTRYID;
    else if (t == ROLE_OUTBOX)
         ret = PR_IPM_OUTBOX_ENTRYID;
    else if (t == ROLE_TRASH)
         ret = PR_IPM_WASTEBASKET_ENTRYID;
    else if (t == ROLE_DRAFT)
         ret = PR_CE_IPM_DRAFTS_ENTRYID;
    else if (t == ROLE_SENT)
         ret = PR_IPM_SENTMAIL_ENTRYID;

    return ret;
}



wstring MessagingManager::getIdOfFolder(const StringBuffer& parentKey, const StringBuffer& folderRole) {
    
    HRESULT hr;
    wstring ret;
    WCHAR* key = toWideChar(parentKey);
    IMsgStore* pStore = getIMsgStore(key);
    if (!pStore) {
        goto FuncExit;
    }
         
    ULONG rgTags[2];
    rgTags[0] = 1;
    rgTags[1] = getFolderEntryId(folderRole);

    ULONG        cValues =  0;
    SPropValue*  rgprops =  NULL;
    LPMAPIFOLDER folder  =  NULL;

     // Now get the desired folder.
    hr = pStore->GetProps ((SPropTagArray *) rgTags, MAPI_UNICODE, &cValues, &rgprops);
    EXIT_ON_FAILED (hr);

    ret = createKeyFromId(rgprops[0].Value.bin); 

FuncExit:
    RELEASE_OBJ(pStore);
    if (key) delete [] key;
    return ret;
}



MailClientData* MessagingManager::completeObject(const ItemData& item, const bool isUpdate) { 

    const WCHAR* mailID = item.getId();
    IMessage* pMsg = getIMessage(mailID);
    
    if (pMsg) {
        MailClientData* mail = new MailClientData(mailID);
        mail->setParent    (item.getParentId());
        mail->setParentRole(item.getParentRole());
 
        // Fills the MailClientData object
        int ret = mailClient2Server.completeObject(*mail, pMsg, isUpdate);
        if (ret) {
            LOG.error("Error in completeObject (code %d)", ret);
        }

        RELEASE_OBJ(pMsg);
        return mail;
    }
    else {
        return NULL;
    }
}


int MessagingManager::moveMessageFromOutboxToSent(const WCHAR* messageID) {

    HRESULT hr = S_OK;
    IMAPIFolder* sentFolder   = NULL;
    IMAPIFolder* sourceFolder = NULL;
    StringBuffer sourceFolderRole;
    wstring      sourceFolderID;

    if (!messageID) {
        return S_FALSE;
    }

    // open the session if necessary
    if (!pSession) {
        hr = logon();
        EXIT_ON_FAILED(hr);
    }


    // Open the message
    IMessage* pMsg = getIMessage(messageID);
    if (!pMsg) {
        LOG.debug("Cannot move mail to sent folder: item with ID = '%S' not found", messageID);
        hr = S_FALSE;  goto FuncExit;
    }

    // Check if the parent folder is an 'outbox'
    sourceFolder = getParentFolder(pMsg);
    if (!sourceFolder) {
        LOG.debug("Cannot move mail %S to sent folder: parent folder not found", messageID);
        hr = S_FALSE;  goto FuncExit;
    }

    sourceFolderID = getEntryIdOfFolder(sourceFolder);
    sourceFolderRole = getFolderRole(sourceFolderID);      // TODO: fix, optimize! (getFolderRole passing the IMAPIFolder*)
    if (sourceFolderRole != ROLE_OUTBOX) {
        LOG.debug("Cannot move mail %S to sent folder: mail is in %s, not in outbox", messageID, sourceFolderRole.c_str());
        hr = S_FALSE;  goto FuncExit;
    }

    //
    // Scan accounts: get the one this email belongs to
    //
    int numAccounts = getAccountNumber();
    for (int i=0; i<numAccounts; i++) {
         StringBuffer accountID = getIdOfAccount(i);
         wstring outboxFolderID = getIdOfFolder(accountID, ROLE_OUTBOX);
         
         if (outboxFolderID == sourceFolderID) {
            // Found!
            // Mark the message as read
            hr = mailClient2Server.markAsRead(pMsg);
            EXIT_ON_FAILED(hr);

            //
            // Move the message
            //
            wstring sentFolderID = getIdOfFolder(accountID, ROLE_SENT);
            sentFolder = getIMapiFolder(sentFolderID.c_str());
            if (!sentFolder) {
                hr = S_FALSE;  goto FuncExit;
            }

            hr = moveMessage(messageID, sourceFolder, sentFolder);
         }
    }

FuncExit:
    RELEASE_OBJ(pMsg);
    RELEASE_OBJ(sourceFolder);
    RELEASE_OBJ(sentFolder);
    return hr;
}



HRESULT MessagingManager::moveMessage(const WCHAR* messageID, IMAPIFolder* sourceFolder, IMAPIFolder* destFolder) {

    if (!messageID || !sourceFolder || !destFolder) {
        return S_FALSE;
    }
    
    ENTRYLIST entryList;
    SBinary mailId = createIdFromKey(messageID);
    entryList.cValues = 1;
    entryList.lpbin = &mailId;    

    // Move the message
    HRESULT hr = sourceFolder->CopyMessages(
                    &entryList,
                    NULL,
                    destFolder,
                    0,
                    NULL,
                    MESSAGE_MOVE);

    return hr;
}



int MessagingManager::sendLog(const StringBuffer& content) {

    // Create the email message
    MailClientData* mcd = mailClient2Server.createMailForSendLog(content);
    if (mcd == NULL) {
        LOG.error("Error creating the SendLog mail");
        return 1;
    }

    // Add the email in the outbox folder of the first available account
    StringBuffer accountID = getIdOfFirstAccount();
    wstring folderID = getIdOfFolder(accountID, ROLE_OUTBOX);
    mcd->setParent(folderID.c_str());
    mcd->setParentRole(ROLE_OUTBOX);
    int ret = addEmailWithAttachment(mcd);

    if (mcd) delete mcd;
    return ret;
}


StringBuffer MessagingManager::getFolderRoleFromLetter(const wstring& folderLetter) {

    if (folderLetter == TEXT("I")) {
        return ROLE_INBOX;
    }
    else if (folderLetter == TEXT("O")) {
        return ROLE_OUTBOX;
    }
    else if (folderLetter == TEXT("S")) {
        return ROLE_SENT;
    }
    else if (folderLetter == TEXT("T")) {
        return ROLE_TRASH;
    }
    else if (folderLetter == TEXT("D")) {
        return ROLE_DRAFT;
    }

    // default
    return ROLE_INBOX;
}


StringBuffer MessagingManager::getDefaultAccountID() {

    StringBuffer defaultName;
    defaultName.convert(DEFAULT_ACCOUNT_NAME);

    StringBuffer id = getIdOfAccount(defaultName);
    
    if (id.empty()) {
        // No default account found!
        // Create here the account for backward compatibility.
        id = createDefaultAccount();
        if (id.empty()) {
            LOG.error("Failed to create default account");
        }
    }
    return id;
}

StringBuffer MessagingManager::getAccountByIndex(unsigned ind)
{
    return getIdOfAccount(ind);
}


StringBuffer MessagingManager::createDefaultAccount() {

    StringBuffer defaultName;
    defaultName.convert(DEFAULT_ACCOUNT_NAME);  // defined in customization.h

    StringBuffer id = getIdOfAccount(defaultName);
    if (id.empty()) {
        // Default account not existing: ok to create it
        MailAccount account;
        account.setName(defaultName);
        account.setVisibleName("");
        account.setEmailAddress(DEFAULT_EMAIL_ADDRESS);

        int ret = createAccount(account);
        if (ret) {
            LOG.error("Error creating account '%s': code %d", defaultName.c_str(), ret);
            return "";
        }

        StringBuffer accountID;
        accountID.convert(account.getID());
        return accountID;
    }
    else {
        // account already existing
        return id;
    }
}



StringBuffer MessagingManager::getFolderRole(const wstring& folderID) {
    
    StringBuffer folderRole = "";

    if (folderID.empty()) {
        return folderRole;
    }

    //
    // Scan accounts and check inbox and outbox folder id   
    //
    int numAccounts = getAccountNumber();
    for (int i=0; i<numAccounts; i++) {
         StringBuffer accountID = getIdOfAccount(i);
         wstring targetFolderID;
         
         targetFolderID = getIdOfFolder(accountID, ROLE_OUTBOX);
         if (folderID == targetFolderID) {
             folderRole = ROLE_OUTBOX;
             break;
         }

         targetFolderID = getIdOfFolder(accountID, ROLE_INBOX);
         if (folderID == targetFolderID) {
             folderRole = ROLE_INBOX;
             break;
         }

         targetFolderID = getIdOfFolder(accountID, ROLE_SENT);
         if (folderID == targetFolderID) {
             folderRole = ROLE_SENT;
             break;
         }
    }

    return folderRole;
}

void MessagingManager::checkMailAccounts()
{
    int numAccounts = getAccountNumber();

    for (int i=0; i < numAccounts; i++) {
        StringBuffer accountId = getIdOfAccount(i);
        const wchar_t* wAccountId = NULL;

        if (accountId.empty()) {
            LOG.error("invalid account ID stored in config");
            continue;
        }

        wAccountId = toWideChar(accountId.c_str());

        // check if account has been removed by user
        if (checkAccount(wAccountId) == false) {
            LOG.debug("account %S not found: removing from client config",
                wAccountId);
            // mark and remove account from client config
            if (markDeleteAccountOnConfig(wAccountId)) {
                LOG.error("error purging config account from config");
            }
            
            // store config
            getRegConfig()->save();
        }

        delete [] wAccountId;
    }
}


bool MessagingManager::checkAccount(const wchar_t* accountId)
{
    bool ret = false;

    if (accountId) {
        unsigned long type = 0;
        LPUNKNOWN* object = openObject(accountId, &type);

        if (object) {
            if (type == MAPI_STORE) {
                ret = true;
            }

            releaseObject(object, type);
        } else {
            LOG.error("error getting object type");
        }
    }

    return ret;
}

unsigned long MessagingManager::getObjectType(const wchar_t* objectId)
{
    unsigned long type = 0L;

    if (objectId) {
        LPUNKNOWN* object = openObject(objectId, &type);

        if (object) {
            releaseObject(object, type);
        } else {
            LOG.error("%s: error getting object type", __FUNCTION__);
        }
    }

    return type;
}
